import argparse

from pix import utils


# fmt: off
def get_parser():
    manifest = utils.get_manifest()

    parser = argparse.ArgumentParser(description=f"{manifest['description']} ({manifest['version']})")
    parser.add_argument("-v", "--version", action="version", version=f"%(prog)s {manifest['version']}")
    subparser = parser.add_subparsers(dest="command", help="Command to execute")

    # convert command
    convert_parser = subparser.add_parser("convert", help="Convert images to a different format")
    convert_parser.add_argument("input", type=str, help="Input image or directory")
    convert_parser.add_argument("-o", "--output", type=str, help="Output image or directory (default: .)", default=".")
    convert_parser.add_argument("-f","--format",type=str,choices=utils.get_supported_formats(),help=f"Output format [{utils.get_supported_formats(True)}]",)
    convert_parser.add_argument("-q", "--quality", type=int, help="Output quality (0-100)", default=95)
    convert_parser.add_argument("--no-optimize", action="store_true", help="Disable optimization")
    convert_parser.add_argument("--overwrite", action="store_true", help="Overwrite existing files")
    convert_parser.add_argument("--prefix", type=str, help="Prefix for the output file name", default="")
    convert_parser.add_argument("--surfix", type=str, help="Surfix for the output file name", default="")

    # resize command
    resize_parser = subparser.add_parser("resize", help="Resize images")
    resize_parser.add_argument("input", type=str, help="Input image or directory")
    resize_parser.add_argument("-o", "--output", type=str, help="Output image or directory", default=".")
    resize_parser.add_argument("--overwrite", action="store_true", help="Overwrite existing files")
    resize_group = resize_parser.add_mutually_exclusive_group()
    resize_group.add_argument("--size", type=str, help="Output size (WxH)")
    resize_group.add_argument("--scale", type=float, help="Output scale (0.0-1.0)")

    # crop command
    crop_parser = subparser.add_parser("crop", help="Crop images")
    crop_parser.add_argument("input", type=str, help="Input image or directory")
    crop_parser.add_argument("-o", "--output", type=str, help="Output image or directory", default=".")
    crop_parser.add_argument("--overwrite", action="store_true", help="Overwrite existing files")
    crop_group = crop_parser.add_mutually_exclusive_group()
    crop_group.add_argument("--ratio", type=str, help="Aspect ratio (W:H). Combine with --align")
    crop_group.add_argument("--size", type=str, help="Exact size and position as WxH+X+Y (e.g., 800x600+100+150)",)
    crop_parser.add_argument("--align", type=str, choices=["top", "bottom", "left", "right", "center"], help="Alignment (top, bottom, left, right, center)")

    # prune command
    prune_parser = subparser.add_parser("prune", help="Remove images smaller than specified resolution")
    prune_parser.add_argument("input", type=str, help="Input image or directory")
    prune_parser.add_argument("-r", "--resolution", type=str, help="Minimum resolution as WxH. Images smaller than this will be removed.")
    prune_parser.add_argument("--dry-run", action="store_true", help="Perform a dry run without deleting files.")
    return parser
# fmt: on



import json
import pathlib


def get_supported_formats(str=False):

    supported_formats = ["png", "jpeg", "webp", "avif", "heif", "tiff", "bmp", "ico"]
    if str:
        return ", ".join(supported_formats)
    else:
        return supported_formats


def is_file_supported(file):
    supported_formats = get_supported_formats()
    ext = pathlib.Path(file).suffix
    ext = ext[1:]
    if ext == "jpg":
        ext = "jpeg"
    return ext in supported_formats


def get_manifest():
    root_dir = pathlib.Path(__file__).parent.absolute()
    manifest_path = pathlib.Path.joinpath(root_dir, "manifest.json")
    with open(manifest_path, "r") as f:
        return json.load(f)



from PIL import Image


def prune_image(input_file, resolution, dry_run=False):
    size = (0, 0)
    with Image.open(input_file) as im:
        size = im.size

    res = tuple(map(int, resolution.split("x")))

    if size < res:
        if dry_run:
            print(f"Would have removed {input_file}")
        else:
            input_file.unlink()
            print(f"Removed {input_file}")


def prune_images(input_dir, resolution, dry_run=False):
    for file in input_dir.iterdir():
        if file.is_file():
            prune_image(file, resolution, dry_run)



import pathlib

from PIL import Image

from pix.utils import is_file_supported


def _save_image(image, output_path, overwrite=False):
    """
    Saves the image to the specified path, with an option to overwrite existing files.
    """
    if not overwrite and pathlib.Path(output_path).exists():
        print(f"File '{output_path}' already exists. Use `--overwrite` to replace it.")
        return
    image.save(output_path)
    print(f"Image saved to '{output_path}'.")


def resize_image(input_path, output_path, size=None, scale=None, overwrite=False):
    """
    Resizes an image based on a specified size or scale factor.
    """
    output_path.parent.mkdir(parents=True, exist_ok=True)

    with Image.open(input_path) as img:
        if size:
            new_size = tuple(map(int, size.split("x")))
        elif scale:
            new_size = tuple([int(dim * scale) for dim in img.size])
        else:
            raise ValueError("Either size or scale must be specified for resizing.")

        resized_img = img.resize(new_size, resample=Image.NEAREST)

        _save_image(resized_img, output_path, overwrite)


def resize_images(input_dir, output_dir, size=None, scale=None, overwrite=False):
    """
    Resizes a directory of images based on a specified size or scale factor.
    """
    for file in input_dir.iterdir():
        if file.is_file() and is_file_supported(file):
            output_file = pathlib.Path.joinpath(output_dir, file.name)
            resize_image(file, output_file, size, scale, overwrite)



import pathlib
import shutil

import pillow_avif  # noqa
from PIL import Image
from pillow_heif import register_heif_opener

from pix.utils import is_file_supported


def _validate_target_format(out, target_format):
    """
    Validates that a target format is provided when the output is a directory.
    """
    if not target_format:
        raise ValueError(
            "Format is required when 'out' is a directory. Use '--format' or '-f' option."
        )


def _get_filename_and_extension(file_path):
    """
    Extracts the file name without extension and the file extension.
    """
    path = pathlib.Path(file_path)
    return path.stem, path.suffix[1:]


def convert_file(
    input_file: pathlib.Path,
    output_file: pathlib.Path,
    prefix="",
    surfix="",
    quality=95,
    optimize=True,
    overwrite=False,
):
    """
    Converts an input file to a specified target format and saves it to the specified output location.
    """
    register_heif_opener()

    # Load and convert the image
    im = Image.open(input_file).convert("RGB")
    output_filename, output_ext = _get_filename_and_extension(output_file)
    if output_ext == "jpg":
        output_ext = "jpeg"
    output_path = pathlib.Path.joinpath(
        output_file.parent, prefix + f"{output_filename}" + surfix + f".{output_ext}"
    )
    output_path.parent.mkdir(parents=True, exist_ok=True)

    # Check if the output file already exists
    if output_path.exists() and not overwrite:
        print(f"`{output_path}` already exists. Use `--overwrite` to replace.")
        return

    im.save(output_path, format=output_ext, quality=quality, optimize=optimize)
    print(f"Image saved to {output_path}, quality: {quality}, optimize: {optimize}")


def convert_files(
    input_dir,
    out_dir,
    target_format,
    prefix="",
    surfix="",
    quality=95,
    optimize=True,
    overwrite=False,
):
    """
    Converts a list of input files to a specified target format and saves them to the specified output location.
    """
    register_heif_opener()

    _validate_target_format(out_dir, target_format)
    for file in input_dir.iterdir():
        if file.is_file() and is_file_supported(file):

            input_file_name, input_file_extension = _get_filename_and_extension(file)
            if input_file_extension == f".{target_format}":
                print(
                    f"[WARNING] Skipping {input_file_name} as it is already in the target format."
                )
                shutil.move(file, f"{out_dir}")
                return
            output_file = pathlib.Path(f"{out_dir}/{input_file_name}.{target_format}")
            convert_file(
                file, output_file, prefix, surfix, quality, optimize, overwrite
            )



import json

from setuptools import find_packages, setup

manifest = json.load(open("pix/manifest.json", "r"))
with open("README.md", "r") as f:
    long_description = f.read()
setup(
    name=manifest["name"],
    version=manifest["version"],
    author=manifest["author"],
    description=manifest["description"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=manifest["url"],
    packages=find_packages(),
    include_package_data=True,
    package_data={
        "": [
            "manifest.json",
        ]
    },
    install_requires=["pillow", "pillow-avif-plugin", "pillow-heif"],
    entry_points={
        "console_scripts": [
            "pix = pix.main:main",
        ],
    },
    python_requires=">=3.9",
    # https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 1 - Planning",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Multimedia :: Graphics :: Graphics Conversion",
        "Topic :: Multimedia :: Graphics :: Editors",
        "Topic :: Multimedia :: Graphics :: Editors :: Raster-Based",
    ],
)



from pathlib import Path

from PIL import Image

from pix.utils import is_file_supported


def _calculate_crop_box_ratio(img_width, img_height, target_ratio, align):
    img_ratio = img_width / img_height
    if target_ratio > img_ratio:
        # Crop vertically
        new_height = int(img_width / target_ratio)
        vertical_start = _calculate_alignment_start(img_height, new_height, align)
        return (0, vertical_start, img_width, vertical_start + new_height)
    else:
        # Crop horizontally
        new_width = int(img_height * target_ratio)
        horizontal_start = _calculate_alignment_start(img_width, new_width, align)
        return (horizontal_start, 0, horizontal_start + new_width, img_height)


def _calculate_alignment_start(total_length, new_length, align):
    if align in ["top", "left"]:
        return 0
    elif align in ["bottom", "right"]:
        return total_length - new_length
    else:  # Center or unspecified
        return (total_length - new_length) // 2


def _crop_image_by_ratio(img, ratio, align):
    target_ratio = float(ratio.split(":")[0]) / float(ratio.split(":")[1])
    box = _calculate_crop_box_ratio(img.width, img.height, target_ratio, align)
    return img.crop(box)


def _crop_image_by_size(img, size_str):
    w, h, x, y = map(int, size_str.replace("x", "+").split("+"))
    box = (x, y, x + w, y + h)
    return img.crop(box)


def crop_image(input_path, output_path, overwrite, **crop_options):
    """crop an image based on the specified options

    Args:
        `input_path` (pathlib.Path): input image to crop
        `output_path` (pathlib.Path): output image path
        `overwrite` (bool): overwrite existing file

    Keyword Arguments:
        `ratio` (str): aspect ratio (W:H). Combine with `align`
        `size` (str): exact size and position as WxH+X+Y (e.g., 800x600+100+150)
        `align` (str): alignment (default: "center") (default: "center")

    Raises:
        ValueError: if neither `ratio` nor `size` is specified
    """

    # create directory if it doesn't exist
    output_path.parent.mkdir(parents=True, exist_ok=True)

    img = Image.open(input_path)
    if "ratio" in crop_options and crop_options["ratio"]:
        cropped_img = _crop_image_by_ratio(
            img, crop_options["ratio"], crop_options.get("align", "center")
        )
    elif "size" in crop_options and crop_options["size"]:
        cropped_img = _crop_image_by_size(img, crop_options["size"])
    else:
        raise ValueError("Either `--ratio` or `--size` must be specified")

    if overwrite or not Path(output_path).exists():
        cropped_img.save(output_path)
        print(
            f"Image saved to {output_path}, ratio: {crop_options.get('ratio', 'N/A')}, align: {crop_options.get('align', 'N/A')}, crop_box: {crop_options.get('size', 'N/A')}"
        )
    else:
        print(f"File {output_path} already exists. Use `--overwrite` to replace it.")


def crop_images(input_dir, output_dir, overwrite, **crop_options):
    """crop a list of images based on the specified options

    Args:
        `input_dir` (pathlib.Path): input directory containing images to crop
        `output_dir` (pathlib.Path): output directory to save cropped images
        `overwrite` (bool): overwrite existing files

    Keyword Arguments:
        `ratio` (str): aspect ratio (W:H). Combine with `align`
        `size` (str): exact size and position as WxH+X+Y (e.g., 800x600+100+150)
        `align` (str): alignment (default: "center") (default: "center")
    """
    for file in input_dir.iterdir():
        if file.is_file() and is_file_supported(file):
            output_file = Path(output_dir, file.name)
            crop_image(file, output_file, overwrite, **crop_options)



import pathlib
import traceback

from pix import cli_parser, converter, croper, prune, resizer


def get_io_paths(args):
    file_input = pathlib.Path(args.input)
    file_output = pathlib.Path(args.output)
    return file_input, file_output


def convert_cmd(args):
    file_input, file_output = get_io_paths(args)
    if file_input.is_file():
        converter.convert_file(
            file_input,
            file_output,
            args.prefix,
            args.surfix,
            args.quality,
            not args.no_optimize,
            args.overwrite,
        )
    elif file_input.is_dir():
        converter.convert_files(
            file_input,
            file_output,
            args.format,
            args.prefix,
            args.surfix,
            args.quality,
            not args.no_optimize,
            args.overwrite,
        )
    else:
        raise FileNotFoundError(f"`{args.input}` not found.")


def resize_cmd(args):
    file_input, file_output = get_io_paths(args)
    if file_input.is_file():

        resizer.resize_image(
            file_input,
            file_output,
            args.size,
            args.scale,
            args.overwrite,
        )
    elif file_input.is_dir():
        resizer.resize_images(
            file_input, file_output, args.size, args.scale, args.overwrite
        )
    else:
        raise FileNotFoundError(f"`{args.input}` not found.")


def crop_cmd(args):
    file_input, file_output = get_io_paths(args)
    if file_input.is_file():
        file_output = pathlib.Path(args.output)
        croper.crop_image(
            file_input,
            file_output,
            args.overwrite,
            ratio=args.ratio,
            size=args.size,
            align=args.align,
        )
    elif file_input.is_dir():
        croper.crop_images(
            file_input,
            file_output,
            args.overwrite,
            ratio=args.ratio,
            size=args.size,
            align=args.align,
        )
    else:
        raise FileNotFoundError(f"`{args.input}` not found.")


def prune_cmd(args):
    file_input = pathlib.Path(args.input)
    if file_input.is_file():
        prune.prune_image(file_input, args.resolution, args.dry_run)
    elif file_input.is_dir():
        prune.prune_images(file_input, args.resolution, args.dry_run)
    else:
        raise FileNotFoundError(f"`{args.input}` not found.")


def main():
    try:
        parser = cli_parser.get_parser()
        args = parser.parse_args()
        if args.command is None:
            parser.print_help()
            return

        # Convert command
        if args.command == "convert":
            convert_cmd(args)

        # Resize command
        elif args.command == "resize":
            resize_cmd(args)

        # Crop command
        elif args.command == "crop":
            crop_cmd(args)

        # Prune command
        elif args.command == "prune":
            prune_cmd(args)
        print("Done!")
    except FileNotFoundError as e:
        print("File not found:", e)
        exit(1)
    except FileExistsError as e:
        print(e)
        exit(1)
    except ValueError as e:
        print(e)
        exit(1)
    except Exception as e:
        print(f"Unhandled exception: {e}")
        traceback.print_exc()
        exit(1)


if __name__ == "__main__":
    main()
